import argparse
import logging
import os
import pprint
import random
import sys
from typing import Any, Dict, List, Optional, Tuple

import torch.distributed

logging.basicConfig(
    format="%(asctime)s | %(levelname)s %(name)s %(message)s)))",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    stream=sys.stdout,
)
logger = logging.getLogger("genhpf.train")

import hydra
import numpy as np
import torch
from hydra.core.hydra_config import HydraConfig
from omegaconf import OmegaConf, open_dict

from genhpf import criterions, models
from genhpf.configs import Config
from genhpf.configs.initialize import add_defaults, hydra_init
from genhpf.datasets import load_dataset
from genhpf.loggings import meters, metrics, progress_bar
from genhpf.trainer import Trainer
from genhpf.utils import checkpoint_utils, distributed_utils, utils


def main(cfg: Config) -> None:
    if distributed_utils.is_master(cfg.distributed_training) and "job_logging_cfg" in cfg:
        # make hydra logging work with ddp (see # see https://github.com/facebookresearch/hydra/issues/1126)
        logging.config.dictConfig(OmegaConf.to_container(cfg.job_logging_cfg))

    if cfg.common.debug:
        os.environ["OMP_NUM_THREADS"] = "2"
        os.environ["MKL_NUM_THREADS"] = "2"
        torch.set_num_threads(2)
        torch.set_num_interop_threads(2)
        cfg.optimization.max_epoch = 1

    assert cfg.dataset.batch_size is not None, "batch_size must be specified"
    metrics.reset()

    np.random.seed(cfg.common.seed)
    random.seed(cfg.common.seed)
    utils.set_torch_seed(cfg.common.seed)

    if distributed_utils.is_master(cfg.distributed_training):
        checkpoint_utils.verify_checkpoint_directory(cfg.checkpoint.save_dir)

    # print args
    logger.info(pprint.pformat(dict(cfg)))

    model = models.build_model(cfg.model)
    if cfg.checkpoint.load_checkpoint is not None:
        state_dict = torch.load(cfg.checkpoint.load_checkpoint, map_location="cpu")["model"]
        model.load_state_dict(state_dict, strict=True)
        logger.info(f"loaded model from {cfg.checkpoint.load_checkpoint}")
    criterion = criterions.build_criterion(cfg.criterion)

    logger.info(model)
    logger.info(f"model: {model.__class__.__name__}")
    logger.info(f"criterion: {criterion.__class__.__name__}")
    logger.info(
        "num. shared model params: {:,} (num. trained: {:,})".format(
            sum(p.numel() for p in model.parameters()),
            sum(p.numel() for p in model.parameters() if p.requires_grad),
        )
    )

    datasets = {}
    train_subsets = cfg.dataset.train_subset.split(",")
    if len(train_subsets) > 1:
        assert (
            cfg.dataset.combine_train_subsets
        ), "train_subset contains multiple datasets, but combine_train_subsets is not set"
        datasets["train"] = [("combined-train", load_dataset(cfg.dataset.data, train_subsets, cfg))]
    else:
        datasets["train"] = [(train_subsets[0].strip(), load_dataset(cfg.dataset.data, train_subsets, cfg))]

    if not cfg.dataset.disable_validation and cfg.dataset.valid_subset is not None:
        valid_subsets = cfg.dataset.valid_subset.split(",")
        if cfg.dataset.combine_valid_subsets:
            datasets["valid"] = [("combined-valid", load_dataset(cfg.dataset.data, valid_subsets, cfg))]
        else:
            datasets["valid"] = [
                (subset.strip(), load_dataset(cfg.dataset.data, [subset], cfg)) for subset in valid_subsets
            ]
    if cfg.dataset.test_subset is not None:
        test_subsets = cfg.dataset.test_subset.split(",")
        if cfg.dataset.combine_test_subsets:
            datasets["test"] = [("combined-test", load_dataset(cfg.dataset.data, test_subsets, cfg))]
        else:
            datasets["test"] = [
                (subset.strip(), load_dataset(cfg.dataset.data, [subset], cfg)) for subset in test_subsets
            ]

    trainer = Trainer(cfg, model, criterion)

    logger.info(f"training on {cfg.distributed_training.distributed_world_size} devices (GPUs)")
    logger.info(f"batch size per device = {cfg.dataset.batch_size}")

    max_epoch = cfg.optimization.max_epoch

    train_meter = meters.StopwatchMeter()
    train_meter.start()
    for i in range(1, max_epoch + 1):
        # train for one epoch
        valid_losses, should_stop = train(cfg, trainer, datasets, i)
        if should_stop:
            break
    train_meter.stop()
    logger.info(f"done training in {train_meter.sum:.1f} seconds")


def should_stop_early(cfg: Config, valid_loss: float) -> bool:
    # skip check if no validation was done in the current epoch
    if valid_loss is None:
        return False
    if cfg.checkpoint.patience <= 0:
        return False

    def is_better(a, b):
        return a > b if cfg.checkpoint.maximize_best_checkpoint_metric else a < b

    prev_best = getattr(should_stop_early, "best", None)
    if prev_best is None or is_better(valid_loss, prev_best):
        should_stop_early.best = valid_loss
        should_stop_early.num_runs = 0
        return False
    else:
        should_stop_early.num_runs += 1
        if should_stop_early.num_runs >= cfg.checkpoint.patience:
            logger.info(
                f"early stop since valid performance hasn't improved for " f"{cfg.checkpoint.patience} runs"
            )
            return True
        else:
            return False


@metrics.aggregate("train")
def train(
    cfg: Config,
    trainer: Trainer,
    datasets,
    epoch: int,
) -> Tuple[List[Optional[float]], bool]:
    """Train the model for one epoch and return validation losses."""
    # initialize data iterator
    data_loader, batch_sampler = trainer.get_train_iterator(datasets["train"][0][1])
    if batch_sampler is not None:
        batch_sampler.set_epoch(epoch)

    itr = iter(data_loader)
    progress = progress_bar.progress_bar(
        itr,
        log_format=cfg.common.log_format,
        log_file=cfg.common.log_file,
        log_interval=cfg.common.log_interval,
        epoch=epoch,
        default_log_format=("tqdm" if cfg.common.no_progress_bar else "simple"),
        wandb_project=(
            cfg.common.wandb_project if distributed_utils.is_master(cfg.distributed_training) else None
        ),
        wandb_entity=(
            cfg.common.wandb_entity if distributed_utils.is_master(cfg.distributed_training) else None
        ),
        wandb_run_name=os.environ.get("WANDB_NAME", os.path.basename(cfg.checkpoint.save_dir)),
    )
    progress.update_config(_flatten_config(cfg))

    logger.info(f"begin training epoch {epoch}")

    should_stop = False
    num_updates = trainer.get_num_updates()
    logger.info("Start iterating over samples")
    for i, sample in enumerate(progress):
        with metrics.aggregate("train_inner"):
            log_output = trainer.train_step(sample)

        if log_output is not None:
            # log mid-epoch stats
            num_updates = trainer.get_num_updates()
            if num_updates % cfg.common.log_interval == 0:
                stats = get_training_stats(metrics.get_smoothed_values("train_inner"))
                progress.log(stats, tag="train_inner", step=num_updates)

                # reset mid-epoch stats after each log interval
                # the end-of-epoch stats will still be preserved
                metrics.reset_meters("train_inner")

    valid_losses, should_stop = validate_and_save(cfg, trainer, datasets, epoch)

    # log end-of-epoch stats
    logger.info(f"end of epoch {epoch} (average epoch stats below)")
    stats = get_training_stats(metrics.get_smoothed_values("train"))
    progress.print(stats, tag="train", step=num_updates)

    # reset epoch-level meters
    metrics.reset_meters("train")
    return valid_losses, should_stop


def validate_and_save(
    cfg: Config,
    trainer: Trainer,
    datasets,
    epoch: int,
) -> Tuple[List[Optional[float]], bool]:
    should_stop = False
    if epoch >= cfg.optimization.max_epoch:
        should_stop = True
        logger.info(
            "Stopping training due to " f"num_epochs: {epoch} >= max_epochs: {cfg.optimization.max_epoch}"
        )

    do_validate = "valid" in datasets or "test" in datasets

    # validate
    valid_losses = [None]
    if do_validate:
        valid_losses = validate(cfg, trainer, datasets, epoch)

    should_stop |= should_stop_early(cfg, valid_losses[0])

    checkpoint_utils.save_checkpoint(cfg.checkpoint, trainer, epoch, valid_losses[0])
    if torch.distributed.is_initialized():
        torch.distributed.barrier()

    return valid_losses, should_stop


def validate(
    cfg: Config,
    trainer: Trainer,
    datasets,
    epoch: int,
):
    """Evaluate the model on the validation set(s) and return the losses."""

    valid_subsets = datasets.get("valid", [])
    test_subsets = datasets.get("test", [])

    valid_losses = []
    for subset, dataset in valid_subsets + test_subsets:
        logger.info(f"begin validation on '{subset}' subset")

        # initialize data iterator
        data_loader, _ = trainer.get_valid_iterator(dataset)
        progress = progress_bar.progress_bar(
            data_loader,
            log_format=cfg.common.log_format,
            log_interval=cfg.common.log_interval,
            log_file=cfg.common.log_file,
            epoch=epoch,
            default_log_format=("tqdm" if cfg.common.no_progress_bar else "simple"),
            wandb_project=(
                cfg.common.wandb_project if distributed_utils.is_master(cfg.distributed_training) else None
            ),
            wandb_entity=(
                cfg.common.wandb_entity if distributed_utils.is_master(cfg.distributed_training) else None
            ),
            wandb_run_name=os.environ.get("WANDB_NAME", os.path.basename(cfg.checkpoint.save_dir)),
        )

        # create a new root metrics aggregator so validation metrics
        # don't pollute other aggregators (e.g., train meters)
        with metrics.aggregate(new_root=True) as agg:
            for i, sample in enumerate(progress):
                trainer.valid_step(sample, subset=subset)

        stats = agg.get_smoothed_values()

        if hasattr(trainer.criterion, "post_validate"):
            stats = trainer.criterion.post_validate(
                stats=stats,
                agg=agg,
            )

        # log validation stats
        stats = get_valid_stats(cfg, trainer, subset, stats)

        progress.print(stats, tag=subset, step=trainer.get_num_updates())

        if np.isnan(stats[cfg.checkpoint.best_checkpoint_metric]):
            logger.info(
                f"validation value for {cfg.checkpoint.best_checkpoint_metric} is NaN. "
                "Changed the best checkpoint metric to loss."
            )
            cfg.checkpoint.best_checkpoint_metric = "loss"
            cfg.checkpoint.maximize_best_checkpoint_metric = False

        valid_losses.append(stats[cfg.checkpoint.best_checkpoint_metric])

    return valid_losses


def get_training_stats(stats):
    stats["wall"] = round(metrics.get_meter("default", "wall").elapsed_time, 0)
    return stats


def get_valid_stats(cfg: Config, trainer: Trainer, subset: str, stats: Dict[str, Any]) -> Dict[str, Any]:
    stats["num_updates"] = trainer.get_num_updates()

    if not hasattr(get_valid_stats, "best"):
        get_valid_stats.best = dict()

    prev_best = getattr(get_valid_stats, "best").get(subset, stats[cfg.checkpoint.best_checkpoint_metric])
    best_function = max if cfg.checkpoint.maximize_best_checkpoint_metric else min
    get_valid_stats.best[subset] = best_function(stats[cfg.checkpoint.best_checkpoint_metric], prev_best)

    key = "best_{0}".format(cfg.checkpoint.best_checkpoint_metric)
    stats[key] = get_valid_stats.best[subset]

    return stats


def _flatten_config(cfg: Config):
    config = OmegaConf.to_container(cfg)
    # remove any legacy Namespaces and replace with a single "args"
    namespace = None
    for k, v in list(config.items()):
        if isinstance(v, argparse.Namespace):
            namespace = v
            del config[k]
    if namespace is not None:
        config["args"] = vars(namespace)
    return config


@hydra.main(config_path=os.path.join("..", "configs"), config_name="config")
def hydra_main(cfg: Config) -> None:
    add_defaults(cfg)

    with open_dict(cfg):
        # make hydra logging work with ddp (see # see https://github.com/facebookresearch/hydra/issues/1126)
        cfg.job_logging_cfg = OmegaConf.to_container(HydraConfig.get().job_logging, resolve=True)

    cfg = OmegaConf.create(OmegaConf.to_container(cfg, resolve=True, enum_to_str=True))
    OmegaConf.set_struct(cfg, True)

    distributed_utils.call_main(cfg, main)


def cli_main():
    try:
        from hydra._internal.utils import get_args

        cfg_name = get_args().config_name or "config"
    except Exception:
        logger.warning("Failed to get config name from hydra args")
        cfg_name = "config"
    hydra_init(cfg_name)
    hydra_main()


if __name__ == "__main__":
    cli_main()



import collections
import logging
import os
import re
import traceback

import torch

from genhpf.configs import CheckpointConfig
from genhpf.utils.file_io import PathManager

logger = logging.getLogger(__name__)


def save_checkpoint(cfg: CheckpointConfig, trainer, epoch: int, val_loss: float):
    from genhpf.loggings import meters

    # only one worker should attempt to create the required dir
    if trainer.data_parallel_rank == 0:
        os.makedirs(cfg.save_dir, exist_ok=True)

    prev_best = getattr(save_checkpoint, "best", val_loss)
    if val_loss is not None:
        best_function = max if cfg.maximize_best_checkpoint_metric else min
        save_checkpoint.best = best_function(val_loss, prev_best)

    if cfg.no_save:
        return

    if not trainer.is_data_parallel_master:
        return

    write_timer = meters.StopwatchMeter()
    write_timer.start()

    updates = trainer.get_num_updates()

    logger.info(f"preparing to save checkpoint for epoch {epoch} @ {updates} updates")

    def is_better(a, b):
        return a >= b if cfg.maximize_best_checkpoint_metric else a <= b

    checkpoint_conds = collections.OrderedDict()
    checkpoint_conds["checkpoint_{}.pt".format(epoch)] = epoch % cfg.save_interval == 0
    checkpoint_conds["checkpoint_best.pt"] = val_loss is not None and (
        not hasattr(save_checkpoint, "best") or is_better(val_loss, save_checkpoint.best)
    )

    checkpoint_conds["checkpoint_last.pt"] = not cfg.no_last_checkpoints

    checkpoints = [os.path.join(cfg.save_dir, fn) for fn, cond in checkpoint_conds.items() if cond]
    if len(checkpoints) > 0:
        trainer.save_checkpoint(checkpoints[0])
        for cp in checkpoints[1:]:
            assert PathManager.copy(
                checkpoints[0], cp, overwrite=True
            ), f"Failed to copy {checkpoints[0]} to {cp}"

        write_timer.stop()
        logger.info(
            "Save checkpoint {} (epoch {} @ {} updates, score {}) (writing took {} seconds)".format(
                checkpoints[0], epoch, updates, val_loss, write_timer.sum
            )
        )

    if cfg.keep_last_epochs > 0:
        # remove old epoch checkpoints; checkpoints are sorted in descending order
        checkpoints = checkpoint_paths(cfg.save_dir, pattern=r"checkpoint_(\d+)\.pt")
        for old_chk in checkpoints[cfg.keep_last_epochs :]:
            if os.path.lexists(old_chk):
                os.remove(old_chk)


def torch_persistent_save(obj, filename, async_write: bool = False):
    if async_write:
        with PathManager.opena(filename, "wb") as f:
            _torch_persistent_save(obj, f)
    else:
        if PathManager.supports_rename(filename):
            # do atmoic save
            with PathManager.open(filename + ".tmp", "wb") as f:
                _torch_persistent_save(obj, f)
            PathManager.rename(filename + ".tmp", filename)


def _torch_persistent_save(obj, f):
    if isinstance(f, str):
        with PathManager.open(f, "wb") as h:
            _torch_persistent_save(obj, h)
        return
    for i in range(3):
        try:
            return torch.save(obj, f)
        except Exception:
            if i == 2:
                logger.error(traceback.format_exc())


def checkpoint_paths(path, pattern=r"checkpoint(\d+)\.pt", keep_match=False):
    """Retrives all checkpoints found in `path` directory.

    Checkpoints are identified by matching filename to the specified pattern. If
    the pattern contains groups, the result will be sorted by the first group in
    descending order
    """
    pt_regexp = re.compile(pattern)
    files = PathManager.ls(path)

    entries = []
    for i, f in enumerate(files):
        m = pt_regexp.fullmatch(f)
        if m is not None:
            idx = float(m.group(1)) if len(m.groups()) > 0 else i
            entries.append((idx, m.group(0)))
    if keep_match:
        return [(os.path.join(path, x[1]), x[0]) for x in sorted(entries, reverse=True)]
    else:
        return [os.path.join(path, x[1]) for x in sorted(entries, reverse=True)]


def load_checkpoint_to_cpu(path, load_on_all_ranks=False):
    """Loads a checkpoint to CPU (with upgrading for backward compatibility).

    If doing single-GPU training or if the checkpoint is only being loaded by at
    most one process on each node (current default behavior is for only rank 0
    to read the checkpoint from disk), load_on_all_ranks should be False to
    avoid errors from torch.distributed not having been initialized or
    torch.distributed.barrier() hanging.

    If all processes on each node may be loading the checkpoint
    simultaneously, load_no_all_ranks should be set to True to avoid I/O
    conflicts.

    There's currently no support for > 1 but < all process loading the
    checkpoint on each node.
    """
    local_path = PathManager.get_local_path(path)
    # The locally cached file returned by get_local_path() may be stable for
    # remote files that are periodically updated/overwritten (ex:
    # checkpoint_last.pt) - so we remove the local copy, sync across processes
    # (if needed), and then download a fresh copy.
    if local_path != path and PathManager.path_requires_pathmanager(path):
        try:
            os.remove(local_path)
        except FileNotFoundError:
            # With potentially multiple processes removing the same file, the
            # file being missing is benign (missing_ok isn't available until
            # Python 3.8).
            pass
        if load_on_all_ranks:
            torch.distributed.barrier()
        local_path = PathManager.get_local_path(path)

    with open(local_path, "rb") as f:
        state = torch.load(f, map_location=torch.device("cpu"))

    return state


def verify_checkpoint_directory(save_dir: str) -> None:
    if not os.path.exists(save_dir):
        os.makedirs(save_dir, exist_ok=True)
    temp_file_path = os.path.join(save_dir, "dummy")
    try:
        with open(temp_file_path, "w"):
            pass
    except OSError as e:
        logger.warning("Unable to access checkpoint save directory: {}".format(save_dir))
        raise e
    else:
        os.remove(temp_file_path)



import logging
from dataclasses import dataclass, field

import torch

from genhpf.models import register_model
from genhpf.models.genhpf import GenHPF, GenHPFConfig
from genhpf.modules import GatherLayer
from genhpf.utils import distributed_utils as dist_utils

logger = logging.getLogger(__name__)


@dataclass
class GenHPFSimCLRConfig(GenHPFConfig):
    all_gather: bool = field(
        default=True, metadata={"help": "whether or not to apply all gather across different gpus"}
    )


@register_model("genhpf_simclr", dataclass=GenHPFSimCLRConfig)
class GenHPFSimCLR(GenHPF):
    def __init__(self, cfg: GenHPFSimCLRConfig):
        super().__init__(cfg)

        self.all_gather = cfg.all_gather

    @classmethod
    def build_model(cls, cfg):
        """Build a new model instance."""
        return cls(cfg)

    def get_logits(self, sample, net_output):
        return net_output

    def get_targets(self, sample, net_output):
        return None

    def forward(
        self,
        input_ids: torch.Tensor,
        type_ids: torch.Tensor = None,
        dpe_ids: torch.Tensor = None,
        padding_mask: torch.Tensor = None,
        **kwargs,
    ):
        x, padding_mask = super().forward(
            input_ids=input_ids, type_ids=type_ids, dpe_ids=dpe_ids, padding_mask=padding_mask, **kwargs
        )

        if padding_mask is not None and padding_mask.any():
            x[padding_mask] = 0
        x = torch.div(x.sum(dim=1), (x != 0).sum(dim=1))

        if self.all_gather and dist_utils.get_data_parallel_world_size() > 1:
            x = torch.cat(GatherLayer.apply(x), dim=0)

        return x



import math

import torch
import torch.nn as nn

class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout, max_len):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model)) 
        pe = torch.zeros(1, max_len, d_model)
        pe[0, :, 0::2] = torch.sin(position * div_term)
        pe[0, :, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x): 
        """
        Args:
            x: Tensor, shape [batch_size, seq_len, embedding_dim]
        """

        x = x + self.pe[:, :x.size(1)]
        return self.dropout(x)


from dataclasses import _MISSING_TYPE, dataclass, field
from typing import Any, List, Optional

from omegaconf import MISSING

from genhpf.configs.constants import LOG_FORMAT_CHOICES


@dataclass
class BaseConfig:
    """base configuration class"""

    _name: Optional[str] = None

    @staticmethod
    def name():
        return None

    def _get_all_attributes(self) -> List[str]:
        return [k for k in self.__dataclass_fields__.keys()]

    def _get_meta(self, attribute_name: str, meta: str, default: Optional[Any] = None) -> Any:
        return self.__dataclass_fields__[attribute_name].metadata.get(meta, default)

    def _get_name(self, attribute_name: str) -> str:
        return self.__dataclass_fields__[attribute_name].name

    def _get_default(self, attribute_name: str) -> Any:
        if hasattr(self, attribute_name):
            if str(getattr(self, attribute_name)).startswith("${"):
                return str(getattr(self, attribute_name))
            elif str(self.__dataclass_fields__[attribute_name].default).startswith("${"):
                return str(self.__dataclass_fields__[attribute_name].default)
            elif getattr(self, attribute_name) != self.__dataclass_fields__[attribute_name].default:
                return getattr(self, attribute_name)

        f = self.__dataclass_fields__[attribute_name]
        if not isinstance(f.default_factory, _MISSING_TYPE):
            return f.default_factory()
        return f.default

    def _get_type(self, attribute_name: str) -> Any:
        return self.__dataclass_fields__[attribute_name].type

    def _get_help(self, attribute_name: str) -> Any:
        return self._get_meta(attribute_name, "help")

    def _get_argparse_const(self, attribute_name: str) -> Any:
        return self._get_meta(attribute_name, "argparse_const")

    def _get_argparse_alias(self, attribute_name: str) -> Any:
        return self._get_meta(attribute_name, "argparse_alias")

    def _get_choices(self, attribute_name: str) -> Any:
        return self._get_meta(attribute_name, "choices")


@dataclass
class CommonConfig(BaseConfig):
    debug: bool = field(default=False, metadata={"help": "enable debug mode"})
    no_progress_bar: bool = field(default=False, metadata={"help": "disable progress bar"})
    log_interval: int = field(default=100, metadata={"help": "log progress every N batches"})
    log_format: Optional[LOG_FORMAT_CHOICES] = field(default=None, metadata={"help": "log format to use"})
    log_file: Optional[str] = field(default=None, metadata={"help": "log file to copy metrics to."})
    wandb_project: Optional[str] = field(
        default=None, metadata={"help": "Weights and Biases project name to use for logging"}
    )
    wandb_entity: Optional[str] = field(
        default=None, metadata={"help": "Weights and Biases entity(team) name to use for logging"}
    )
    seed: int = field(default=42, metadata={"help": "random seed"})
    all_gather_list_size: int = field(
        default=32768,
        metadata={"help": "number of bytes reserved for gathering stats from workers"},
    )


@dataclass
class DistributedTrainingConfig(BaseConfig):
    distributed_world_size: int = field(default=1, metadata={"help": "total number of GPUs across all nodes"})
    distributed_rank: Optional[int] = field(default=0, metadata={"help": "rank of the current worker"})
    distributed_backend: str = field(default="nccl", metadata={"help": "distributed backend"})
    distributed_init_method: Optional[str] = field(
        default=None,
        metadata={"help": "typically tcp://hostname:port that will be used to " "init distributed training"},
    )
    distributed_port: int = field(default=12355, metadata={"help": "port number for distributed training"})
    device_id: int = field(default=0, metadata={"help": "which GPU to use"})
    bucket_cap_mb: int = field(default=25, metadata={"help": "bucket size for reduction"})
    find_unused_parameters: bool = field(
        default=False, metadata={"help": "disable unused parameter detection when using distributed training"}
    )
    broadcast_buffers: bool = field(
        default=False,
        metadata={
            "help": "Copy non-trainable parameters between GPUs, such as " "batchnorm population statistics"
        },
    )


@dataclass
class DatasetConfig(BaseConfig):
    data_format: str = field(
        default="genhpf", metadata={"help": "data format. supported formats: genhpf, meds"}
    )
    data: str = field(default=MISSING, metadata={"help": "path to the data directory"})
    label: bool = field(default=False, metadata={"help": "whether to load labels from the dataset"})
    vocab_size: int = field(default=MISSING, metadata={"help": "size of the vocabulary"})
    pad_token_id: int = field(default=0, metadata={"help": "pad token id"})
    sep_token_id: int = field(default=102, metadata={"help": "sep token id"})
    dummy_token_id: int = field(default=101, metadata={"help": "dummy token id"})
    ignore_index: int = field(
        default=-100,
        metadata={
            "help": "specifies a target value that is ignored and does not contribute to "
            "the input gradient. only applied to cross-entropy loss"
        },
    )
    apply_mask: bool = field(default=False, metadata={"help": "whether to apply masking to the input tokens"})
    mask_prob: float = field(default=0.15, metadata={"help": "probability for masking tokens"})
    mask_unit: str = field(
        default="token", metadata={"help": "unit for masking. supported units: token, event"}
    )
    mask_token_id: int = field(default=103, metadata={"help": "mask token id"})
    num_workers: int = field(default=1, metadata={"help": "how many subprocesses to use for data loading"})
    batch_size: Optional[int] = field(default=None, metadata={"help": "number of examples in a batch"})
    train_subset: str = field(
        default="train", metadata={"help": "data subset name to use for training (e.g., train, valid, test)"}
    )
    valid_subset: Optional[str] = field(
        default="valid", metadata={"help": "comma separated list of data subset names to use for validation"}
    )
    test_subset: Optional[str] = field(
        default="test", metadata={"help": "comma separated list of data subset names to use for test"}
    )
    combine_train_subsets: Optional[bool] = field(
        default=None,
        metadata={
            "help": "whether to combine all training subsets into one dataset",
            "argparse_alias": "--combine-train",
        },
    )
    combine_valid_subsets: Optional[bool] = field(
        default=None,
        metadata={
            "help": "whether to combine all validation subsets into one dataset",
            "argparse_alias": "--combine-val",
        },
    )
    combine_test_subsets: Optional[bool] = field(
        default=None,
        metadata={
            "help": "whether to combine all test subsets into one dataset",
            "argparse_alias": "--combine-test",
        },
    )
    disable_validation: Optional[bool] = field(
        default=None, metadata={"help": "whether to disable validation during training"}
    )


@dataclass
class OptimizationConfig(BaseConfig):
    max_epoch: int = field(default=0, metadata={"help": "maximum number of epochs to train"})
    lr: float = field(default=1e-4, metadata={"help": "learning rate"})
    adam_betas: Any = field(default=(0.9, 0.999), metadata={"help": "betas for Adam optimizer"})
    adam_eps: float = field(default=1e-8, metadata={"help": "epsilon for Adam optimizer"})
    weight_decay: float = field(default=1e-8, metadata={"help": "weight decay for Adam optimizer"})


@dataclass
class CheckpointConfig(BaseConfig):
    save_dir: str = field(default="checkpoints", metadata={"help": "path to save checkpoints"})
    checkpoint_prefix: str = field(
        default="checkpoint", metadata={"help": "prefix to add to the checkpoint file name"}
    )
    checkpoint_suffix: str = field(default="", metadata={"help": "suffix to add to the checkpoint file name"})
    load_checkpoint: Optional[str] = field(
        default=None, metadata={"help": "path to a checkpoint to load model weights from, if provided"}
    )
    no_save: bool = field(default=False, metadata={"help": "don't save checkpoints"})
    save_interval: int = field(default=1, metadata={"help": "save a checkpoint every N epochs"})
    keep_last_epochs: int = field(default=-1, metadata={"help": "keep last N epoch checkpoints"})
    no_last_checkpoints: bool = field(default=False, metadata={"help": "don't store last checkpoints"})
    best_checkpoint_metric: str = field(
        default="loss", metadata={"help": 'metric to use for saving "best" checkpoints'}
    )
    maximize_best_checkpoint_metric: bool = field(
        default=False,
        metadata={"help": 'select the largest metric value for saving "best" checkpoints'},
    )
    patience: int = field(
        default=-1,
        metadata={
            "help": (
                "early stop training if valid performance doesn't "
                "improve for N consecutive validation runs; note "
                "that this is influenced by --validate-interval"
            )
        },
    )


@dataclass
class MEDSConfig(BaseConfig):
    output_predictions: bool = field(
        default=False,
        metadata={
            "help": "whether to output predictions. if turned on, `genhpf-test` will automatically "
            "output predictions of the test set specified by `dataset.test_subset` in the "
            "`meds.output_dir` directory."
        },
    )
    labels_dir: Optional[str] = field(
        default=None,
        metadata={
            "help": "a path to the label directory for MEDS dataset. this is required to store "
            "output predictions in the format expected by `meds-evaluation` when "
            "`meds.output_predictions` is turned on."
        },
    )
    output_dir: Optional[str] = field(
        default=None,
        metadata={
            "help": "a path to the output directory to store output predictions. "
            "this is only used when `meds.output_predictions` is turned on."
        },
    )


@dataclass
class Config(BaseConfig):
    common: CommonConfig = field(default_factory=CommonConfig)
    distributed_training: DistributedTrainingConfig = field(default_factory=DistributedTrainingConfig)
    dataset: DatasetConfig = field(default_factory=DatasetConfig)
    optimization: OptimizationConfig = field(default_factory=OptimizationConfig)
    checkpoint: CheckpointConfig = field(default_factory=CheckpointConfig)
    meds: MEDSConfig = field(default_factory=MEDSConfig)
    model: Any = MISSING
    criterion: Any = None



import logging
from dataclasses import dataclass

import torch
import torch.nn as nn
from omegaconf import II

from genhpf.models import register_model
from genhpf.models.genhpf import GenHPF, GenHPFConfig

logger = logging.getLogger(__name__)


@dataclass
class GenHPFMLMConfig(GenHPFConfig):
    ignore_index: int = II("dataset.ignore_index")


@register_model("genhpf_mlm", dataclass=GenHPFMLMConfig)
class GenHPFMLM(GenHPF):
    def __init__(self, cfg: GenHPFMLMConfig):
        super().__init__(cfg)

        self.ignore_index = cfg.ignore_index

        self.input_ids_proj = nn.Linear(cfg.agg_embed_dim, cfg.vocab_size)

    @classmethod
    def build_model(cls, cfg):
        """Build a new model instance."""
        return cls(cfg)

    def get_logits(self, sample, net_output):
        masked_indices = torch.where(
            (sample["input_label"] > 0) & (sample["input_label"] != self.ignore_index)
        )
        return net_output["input_ids"][masked_indices]

    def get_targets(self, sample, net_output):
        masked_indices = torch.where(
            (sample["input_label"] > 0) & (sample["input_label"] != self.ignore_index)
        )
        return sample["input_label"][masked_indices]

    def forward(
        self,
        input_ids: torch.Tensor,
        type_ids: torch.Tensor = None,
        dpe_ids: torch.Tensor = None,
        padding_mask: torch.Tensor = None,
        **kwargs,
    ):
        x, padding_mask = super().forward(
            input_ids=input_ids, type_ids=type_ids, dpe_ids=dpe_ids, padding_mask=padding_mask, **kwargs
        )

        input_ids = self.input_ids_proj(x)

        return {"input_ids": input_ids, "padding_mask": padding_mask}

    def get_pretraining_parameter_names(self):
        ret = []
        ret.extend(["input_ids_proj" + "." + x[0] for x in self.input_ids_proj.named_parameters()])
        return ret



import math
import re
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import genhpf.utils.utils as utils
from genhpf.criterions import BaseCriterion, register_criterion
from genhpf.criterions.criterion import CriterionConfig
from genhpf.loggings import metrics
from genhpf.models.genhpf import GenHPF

from . import build_criterion


@dataclass
class MultiTaskCriterionConfig(CriterionConfig):
    task_loss_weights: Optional[List[float]] = field(
        default=None,
        metadata={
            "help": "weights for each loss term. if given, has to be a float list of size " "n_criterions"
        },
    )
    args: Any = field(
        default=None,
        metadata={
            "help": "configurations for each criterion where the name of each argument should "
            "match with the corresponding task name."
        },
    )


@register_criterion("multi_task_criterion", dataclass=MultiTaskCriterionConfig)
class MultiTaskCriterion(BaseCriterion):
    def __init__(self, cfg: MultiTaskCriterionConfig):
        super().__init__(cfg)

        criterions = {}
        for task_name in self.task_names:
            criterion_cfg = getattr(cfg.args, task_name)
            criterions[task_name] = build_criterion(criterion_cfg)
        self.criterions = criterions

        if cfg.task_loss_weights is None:
            self.task_loss_weights = [1.0] * len(criterions)
        else:
            self.task_loss_weights = cfg.task_loss_weights

    def forward(self, model: GenHPF, sample, return_net_output=False):
        net_output = model(**sample["net_input"])
        logits = model.get_logits(sample, net_output)
        targets = model.get_targets(sample, net_output)

        if not isinstance(logits, dict):
            logits = {self.task_names[0]: logits}
        if not isinstance(targets, dict):
            targets = {self.task_names[0]: targets}

        if len(logits) != len(self.task_names) or len(targets) != len(self.task_names):
            raise ValueError(
                "number of logits and targets should be equal to the number of tasks. "
                f"got {len(logits)} logits and {len(targets)} targets for "
                f"{len(self.task_names)} tasks"
            )

        loss = 0.0
        logging_outputs = dict()
        for i, task_name in enumerate(self.task_names):
            criterion = self.criterions[task_name]
            assert (
                task_name in logits and task_name in targets
            ), f"task name {task_name} not found in logits or targets"
            task_logits = logits[task_name]
            task_targets = targets[task_name]
            task_loss, task_losses_to_log = criterion.compute_loss(
                logits=task_logits, targets=task_targets, sample=sample, net_output=net_output, model=model
            )
            task_loss *= self.task_loss_weights[i]
            sample_size = criterion.get_sample_size(sample, task_targets)

            logging_outputs[f"<{task_name}>_criterion_cls"] = criterion.__class__
            if len(task_losses_to_log) > 1:
                logging_outputs[f"{task_name}_loss"] = task_loss.item()
                for j, l in enumerate(task_losses_to_log):
                    logging_outputs[f"<{task_name}>_loss_{j}"] = l
            else:
                logging_outputs[f"<{task_name}>_loss"] = task_losses_to_log[0]
            logging_outputs[f"<{task_name}>_sample_size"] = sample_size

            task_logging_output = criterion.get_logging_outputs({}, task_logits, task_targets, sample)
            for log, value in task_logging_output.items():
                if log.startswith("_"):
                    log = log[1:]
                    logging_outputs[f"_<{task_name}>_{log}"] = value
                else:
                    logging_outputs[f"<{task_name}>_{log}"] = value

            # divide task loss by the sample size beforehand to handle different sample
            # sizes for multiple criterions
            loss += task_loss / logging_outputs[f"<{task_name}>_sample_size"]

        # manipulate sample_size to be 1 to avoid double-dividing gradients in optimizer later
        sample_size = 1

        if return_net_output:
            return loss, sample_size, logging_outputs, net_output
        else:
            return loss, sample_size, logging_outputs

    @staticmethod
    def reduce_metrics(logging_outputs: List[Dict[str, Any]]) -> None:
        log_keys = logging_outputs[0].keys()

        grouped_log_keys = defaultdict(list)
        for lk in log_keys:
            group = re.search(r"\<.*\>", lk)
            offset = group.end() + 1
            group = group.group()[1:-1]
            key = lk[offset:]
            if lk.startswith("_"):
                key = "_" + key
            grouped_log_keys[group].append(key)

        total_loss = 0
        for group, log_keys in grouped_log_keys.items():
            criterion_cls = logging_outputs[0][f"<{group}>_criterion_cls"]
            logging_output = []
            for log in logging_outputs:
                log_dict = {}
                for log_key in set(log_keys) - {"criterion_cls"}:
                    if log_key.startswith("_") and f"_<{group}>{log_key}" in log:
                        log_dict[log_key] = log[f"_<{group}>{log_key}"]
                    elif f"<{group}>_{log_key}" in log:
                        log_dict[log_key] = log[f"<{group}>_{log_key}"]
                logging_output.append(log_dict)
            criterion_cls.reduce_metrics(logging_output, prefix=group)

            loss_sum = utils.item(sum(log.get("loss", 0) for log in logging_output))
            sample_size = utils.item(sum(log.get("sample_size", 0) for log in logging_output))

            total_loss += loss_sum / (sample_size or 1) / math.log(2)

        metrics.log_scalar("loss", total_loss, 1, round=3)

    def post_validate(self, stats, agg, **kwargs):
        task_agg = {}
        for key in agg:
            for task_name in self.task_names:
                if key.startswith(task_name) or key[1:].startswith(task_name):
                    if task_name not in task_agg:
                        task_agg[task_name] = {}
                    task_agg[task_name][key] = agg[key]
                    break

        for task_name, task_agg in task_agg.items():
            if hasattr(self.criterions[task_name], "post_validate"):
                stats = self.criterions[task_name].post_validate(stats, task_agg, **kwargs)

        for key in list(stats.keys()):
            for task_name in self.task_names:
                if key.startswith(task_name):
                    stat_key = key[len(task_name) + 1 :]
                    if f"avg_{stat_key}" not in stats:
                        stats[f"avg_{stat_key}"] = []
                    stats[f"avg_{stat_key}"].append(stats[key])
                    break

        for key in list(stats.keys()):
            if key.startswith("avg_"):
                stats[key] = sum(stats[key]) / len(stats[key])
        return stats

    def eval(self):
        super().eval()
        for criterion in self.criterions.values():
            criterion.eval()
        return self



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import bisect
import time
from collections import OrderedDict
from typing import Dict, Optional

import torch
import numpy as np
from sklearn.metrics import average_precision_score, roc_auc_score

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

def type_as(a, b):
    if torch.is_tensor(a) and torch.is_tensor(b):
        return a.to(b)
    else:
        return a

class Meter(object):
    """Base class for Meters."""

    def __init__(self):
        pass

    def state_dict(self):
        return {}
    
    def load_state_dict(self, state_dict):
        pass

    def reset(self):
        raise NotImplementedError
    
    @property
    def smoothed_value(self) -> float:
        """Smoothed value used for logging."""
        raise NotImplementedError

def safe_round(number, ndigits):
    if hasattr(number, "__round__"):
        return round(number, ndigits)
    elif torch.is_tensor(number) and number.numel() == 1:
        return safe_round(number.item(), ndigits)
    elif np.ndim(number) == 0 and hasattr(number, "item"):
        return safe_round(number.item(), ndigits)
    else:
        return number

class SumMeter(Meter):
    """Computes and stores the sum"""

    def __init__(self, round: Optional[int] = None):
        self.round = round
        self.reset()

    def reset(self):
        self.sum = 0  # sum from all updates

    def update(self, val):
        if val is not None:
            self.sum = type_as(self.sum, val) + val

    def state_dict(self):
        return {
            "sum": self.sum,
            "round": self.round,
        }

    def load_state_dict(self, state_dict):
        self.sum = state_dict["sum"]
        self.round = state_dict.get("round", None)

    @property
    def smoothed_value(self) -> float:
        val = self.sum
        if self.round is not None and val is not None:
            val = safe_round(val, self.round)
        return val

class AverageMeter(Meter):
    """Computes and stores the average and current value"""

    def __init__(self, round: Optional[int] = None):
        self.round = round
        self.reset()
    
    def reset(self):
        self.val = None # most recent update
        self.sum = 0 # sum from all updates
        self.count = 0 # total n from all updates
    
    def update(self, val, n = 1):
        if val is not None:
            self.val = val
            if n > 0:
                self.sum = type_as(self.sum, val) + (val * n)
                self.count = type_as(self.count, n) + n
    
    def state_dict(self):
        return {
            "val" : self.val,
            "sum" : self.sum,
            "count" : self.count,
            "round" : self.round
        }
    
    def load_state_dict(self, state_dict):
        self.val = state_dict["val"]
        self.sum = state_dict["sum"]
        self.count = state_dict["count"]
        self.round = state_dict.get("round", None)
    
    @property
    def avg(self):
        return self.sum / self.count if self.count > 0 else self.val
    
    @property
    def smoothed_value(self) -> float:
        val = self.avg
        if self.round is not None and val is not None:
            val = safe_round(val, self.round)
        return val

class TimeMeter(Meter):
    """Compute the average occurrence of some event per second"""

    def __init__(
        self,
        init: int = 0,
        n: int = 0,
        round: Optional[int] = None
    ):
        self.round = round
        self.reset(init, n)
    
    def reset(self, init = 0, n = 0):
        self.init = init
        self.start = time.perf_counter()
        self.n = n
        self.i = 0

    def update(self, val = 1):
        self.n = type_as(self.n, val) + val
        self.i += 1
    
    def state_dict(self):
        return {
            "init": self.elapsed_time,
            "n": self.n,
            "round": self.round
        }
    
    def load_state_dict(self, state_dict):
        if "start" in state_dict:
            # backwards compatibility for old state_dicts
            self.reset(init = state_dict["init"])
        else:
            self.reset(init = state_dict["init"], n = state_dict["n"])
            self.round = state_dict.get("round", None)
    
    @property
    def avg(self):
        return self.n / self.elapsed_time

    @property
    def elapsed_time(self):
        return self.init + (time.perf_counter() - self.start)
    
    @property
    def smoothed_value(self) -> float:
        val = self.avg
        if self.round is not None and val is not None:
            val = safe_round(val, self.round)
        return val


class AUCMeter(Meter):
    "Stores scores / targets to compute AUROC and AUPRC"

    def __init__(self,):
        self.reset()
    
    def reset(self):
        self.scores = []
        self.targets = []
    
    def update(self, prob, target):
        if torch.is_tensor(prob):
            prob = prob.cpu().numpy()
        if torch.is_tensor(target):
            target = target.cpu().numpy()

        self.scores.append(prob)
        self.targets.append(target)
    
    def state_dict(self):
        return {
            "scores": self.scores,
            "targets": self.targets,
        }
    
    def load_state_dict(self, state_dict):
        self.scores = state_dict["scores"]
        self.targets = state_dict["targets"]
        self.round = state_dict.get("round", None)

    @property
    def auroc(self):
        y_true = np.concatenate(self.targets)
        y_score = np.concatenate(self.scores)
        # if y_true.shape != y_score.shape:
        #     y_true = np.eye(y_score.shape[1])[y_true]
        if y_true.shape[0] > 127 and len(y_true.shape) >1:
            mask = (y_true.sum(axis=0)!=0)
            y_true = y_true[:, mask]
            y_score = y_score[:, mask]
        try:
            return roc_auc_score(y_true=y_true, y_score=y_score, average='macro')
        except ValueError:
            return float("nan")
    
    @property
    def auprc(self):
        y_true = np.concatenate(self.targets)
        y_score = np.concatenate(self.scores)
        if y_true.shape != y_score.shape:
            y_true = np.eye(y_score.shape[1])[y_true]
        try:
            return average_precision_score(y_true=y_true, y_score=y_score, average='micro')
        except ValueError:
            return float("nan")

    @property
    def smoothed_value(self) -> float:
        raise AttributeError(
            "AUC meter cannot have smoothed values. Please "
            "make sure the key of this meter starts with '_'."
        )
        
    
class StopwatchMeter(Meter):
    """Computes the sum/avg duration of some event in seconds"""

    def __init__(self, round: Optional[int] = None):
        self.round = round
        self.sum = 0
        self.n = 0
        self.start_time = None
    
    def start(self):
        self.start_time = time.perf_counter()
    
    def stop(self, n = 1, prehook = None):
        if self.start_time is not None:
            if prehook is not None:
                prehook()
            delta = time.perf_counter() - self.start_time
            self.sum = self.sum + delta
            self.n = type_as(self.n, n) + n
    
    def reset(self):
        self.sum = 0 # cumulative time during which stopwatch was active
        self.n = 0 # total n across all start/stop
        self.start()
    
    def state_dict(self):
        return {
            "sum": self.sum,
            "n": self.n,
            "round": self.round
        }
    
    def load_state_dict(self, state_dict):
        self.sum = state_dict["sum"]
        self.n = state_dict["n"]
        self.start_time = None
        self.round = state_dict.get("round", None)
    
    @property
    def avg(self):
        return self.sum / self.n if self.n > 0 else self.sum
    
    @property
    def elapsed_time(self):
        if self.start_time is None:
            return 0.0
        return time.perf_counter() - self.start_time
    
    @property
    def smoothed_value(self) -> float:
        val = self.avg if self.sum > 0 else self.elapsed_time
        if self.round is not None and val is not None:
            val = safe_round(val, self.round)
        return val

class MetersDict(OrderedDict):
    """A sorted dictionary of :class:`Meters`.

    Meters are sorted according to a priority that is given when the
    meter is first added to the dictionary.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.priorities = []
    
    def __setitem__(self, key, value):
        assert key not in self, "MetersDict doesn't support reassignment"
        priority, value = value
        bisect.insort(self.priorities, (priority, len(self.priorities), key))
        super().__setitem__(key, value)
        for _, _, key in self.priorities: # reorder dict to match priorities
            self.move_to_end(key)
    
    def add_meter(self, key, meter, priority):
        self.__setitem__(key, (priority, meter))
    
    def state_dict(self):
        return [
            (pri, key, self[key].__class__.__name__, self[key].state_dict())
            for pri, _, key in self.priorities
            # can't serialize DerivedMeter instances
            if not isinstance(self[key], MetersDict._DerivedMeter)
        ]
    
    def load_state_dict(self, state_dict):
        self.clear()
        self.priorities.clear()
        for pri, key, meter_cls, meter_state in state_dict:
            meter = globals()[meter_cls]()
            meter.load_state_dict(meter_state)
            self.add_meter(key, meter, pri)
    
    def get_smoothed_value(self, key: str) -> float:
        """Get a single smoothed value."""
        meter = self[key]
        if isinstance(meter, MetersDict._DerivedMeter):
            # print("hello: ", key, meter.fn(self))
            return meter.fn(self)
        else:
            return meter.smoothed_value
    
    def get_smoothed_values(self) -> Dict[str, float]:
        """Get all smoothed values."""
        return OrderedDict(
            [
                (key, self.get_smoothed_value(key))
                for key in self.keys()
                if not key.startswith("_")
            ]
        )
    
    def reset(self):
        """Reset Meter instances."""
        for meter in self.values():
            if isinstance(meter, MetersDict._DerivedMeter):
                continue
            meter.reset()
    
    class _DerivedMeter(Meter):
        """A Meter whose values are derived from other Meters."""

        def __init__(self, fn):
            self.fn = fn

        def reset(self):
            pass


import importlib
import os

from hydra.core.config_store import ConfigStore

from .genhpf import GenHPF #noqa
from genhpf.configs.utils import merge_with_parent

MODEL_REGISTRY = {}
MODEL_DATACLASS_REGISTRY = {}

def build_model(cfg):
    model = None
    model_type = getattr(cfg, "_name", None)

    if model_type in MODEL_REGISTRY:
        model = MODEL_REGISTRY[model_type]
        # set defaults from dataclass
        dc = MODEL_DATACLASS_REGISTRY[model_type]
        cfg = merge_with_parent(dc(), cfg)

    assert model is not None, (
        f"Could not infer model type from {str(model_type)}. "
        + "Available models: "
        + str(MODEL_REGISTRY.keys())
        + " Requested model type: "
        + str(model_type)
    )

    model_instance = model.build_model(cfg)
    return model_instance

def register_model(name, dataclass=None):
    """
    New model types can be added with the :func:`register_model`
    function decorator.

    Args:
        name (str): the name of the model
    """

    def register_model_cls(cls):
        if name in MODEL_REGISTRY:
            raise ValueError(f"Cannot register duplicate model ({name})")
        if not issubclass(cls, GenHPF):
            raise ValueError(f"Model ({name}: {cls.__name__}) must extend GenHPF")
        MODEL_REGISTRY[name] = cls
        if dataclass is not None:
            MODEL_DATACLASS_REGISTRY[name] = dataclass

            cs = ConfigStore.instance()
            node = dataclass()
            node._name = name
            cs.store(name=name, group="model", node=node, provider="genhpf")

        return cls

    return register_model_cls

def import_models(models_dir, namespace):
    for file in os.listdir(models_dir):
        path = os.path.join(models_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            model_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + model_name)

# automatically import any Python files in the models/ directory
models_dir = os.path.dirname(__file__)
import_models(models_dir, "genhpf.models")


import importlib
import os

from hydra.core.config_store import ConfigStore

from genhpf.criterions.criterion import BaseCriterion #noqa
from genhpf.configs.utils import merge_with_parent

CRITERION_REGISTRY = {}
CRITERION_DATACLASS_REGISTRY = {}

def build_criterion(cfg) -> BaseCriterion:
    criterion = None
    criterion_type = getattr(cfg, "_name", None)

    if criterion_type in CRITERION_REGISTRY:
        criterion = CRITERION_REGISTRY[criterion_type]
        # set defaults from dataclass
        dc = CRITERION_DATACLASS_REGISTRY[criterion_type]
        cfg = merge_with_parent(dc(), cfg)

    assert criterion is not None, (
        f"Could not infer criterion type from {str(criterion_type)}. "
        + "Available criterions: "
        + str(CRITERION_REGISTRY.keys())
        + " Requested criterion type: "
        + str(criterion_type)
    )

    return criterion.build_criterion(cfg)

def register_criterion(name, dataclass=None):
    """
    New criterion types can be added with the :func:`register_criterion`
    function decorator.

    Args:
        name (str): the name of the criterion
    """

    def register_criterion_cls(cls):
        if name in CRITERION_REGISTRY:
            raise ValueError(f"Cannot register duplicate criterion ({name})")
        if not issubclass(cls, BaseCriterion):
            raise ValueError(
                f"Criterion ({name}: {cls.__name__}) must extend Base Criterion"
            )
        CRITERION_REGISTRY[name] = cls
        if dataclass is not None:
            CRITERION_DATACLASS_REGISTRY[name] = dataclass

            cs = ConfigStore.instance()
            node = dataclass()
            node._name = name
            cs.store(name=name, group="criterion", node=node, provider="genhpf")

        return cls

    return register_criterion_cls

def import_criterions(criterions_dir, namespace):
    for file in os.listdir(criterions_dir):
        path = os.path.join(criterions_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            criterion_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + criterion_name)

# automatically import any Python files in the criterions/ directory
criterions_dir = os.path.dirname(__file__)
import_criterions(criterions_dir, "genhpf.criterions")


import io
from typing import Optional, List, Any
import logging
import os
import random
import socket
import struct
import pickle
import warnings
from dataclasses import dataclass

import torch
import torch.nn as nn
import torch.distributed as dist

from torch.nn.parallel import DistributedDataParallel

from genhpf.configs import Config, DistributedTrainingConfig

logger = logging.getLogger(__name__)

class ModuleProxyWrapper(nn.Module):
    """
    Wrap a DistributedDataParallel module and forward requests for missing
    attributes to the module wrapped by DDP (the twice-wrapped module).
    Also forward calls to :func:`state_dict` and :func:`load_state_dict`.
    
    Usage::

        module.xyz = "hello world"
        wrapped_module = DistributedDataParallel(module, **ddp_args)
        wrapped_module = ModuleProxyWrapper(wrapped_module)
        assert wrapped_module.xyz == "hello world"
        assert wrapped_module.state_dict().keys() == module.state_dict().keys()

    Args:
        module (nn.Module): module to wrap
    """

    def __init__(self, module: nn.Module):
        super().__init__()
        assert hasattr(module, "module"), \
            "ModuleProxyWrapper expects input to wrap another module"
        self.module = module
    
    def __getattr__(self, name):
        """Forward missing attributes to twice-wrapped module."""
        try:
            # defer to nn.Module's logic
            return super().__getattr__(name)
        except AttributeError:
            try:
                # forward to the once-wrapped module
                return getattr(self.module, name)
            except AttributeError:
                # forward to the twice-wrapped module
                return getattr(self.module.module, name)
    
    def state_dict(self, *args, **kwargs):
        """Forward to the twice-wrapped module."""
        return self.module.module.state_dict(*args, **kwargs)
    
    def load_state_dict(self, *args, **kwargs):
        """Forward to the twice-wrapped module."""
        return self.module.module.load_state_dict(*args, **kwargs)
    
    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)

def DistributedModel(
    args,
    model,
    process_group,
    device
):
    """
    Wrap a *model* to support distributed data parallel training.

    This is similar to the built-in DistributedDataParallel, but allows
    additional configuration of the DistributedDataParallel class to
    use, and also provides easier access to the wrapped model by
    forwarding requests for missing attributes to the wrapped model.

    Args:
        args (argparse.Namespace): genhpf model args
        model (BaseModel): model to wrap
        process_group: the c10d process group to be used for distributed data
            parallel all-reduction
        device: device to move model to
    """
    assert isinstance(model, nn.Module)
    
    wrapped_model = DistributedDataParallel(
        module = model.to(device),
        device_ids = [args.device_id],
        output_device = args.device_id,
        broadcast_buffers = args.broadcast_buffers,
        bucket_cap_mb = args.bucket_cap_mb,
        process_group = process_group,
        find_unused_parameters = args.find_unused_parameters
    )

    if args.ddp_comm_hook == "fp16":
        logger.info("enable fp16 communication hook in DDP")
        try:
            from torch.distributed.algorithms.ddp_comm_hooks import (
                register_ddp_comm_hook,
                DDPCommHookType,
            )
        except:
            logger.error(
                "Could not import from torch.distributed.algorithms.ddp_comm_hooks; you may need to update your pytorch version"
            )
            raise

        register_ddp_comm_hook(DDPCommHookType.FP16_COMPRESS, wrapped_model)
        
    # forward missing getattr and state_dict/load_state_dict to orig model
    wrapped_model = ModuleProxyWrapper(wrapped_model)
    
    return wrapped_model

def is_master(cfg: DistributedTrainingConfig):
    return cfg.distributed_rank == 0

def infer_init_method(cfg: DistributedTrainingConfig):
    assert (
        cfg.distributed_world_size <= torch.cuda.device_count()
    ), f"world size is {cfg.distributed_world_size} but have {torch.cuda.device_count()} available devices"
    port = random.randint(10000, 20000)
    cfg.distributed_init_method = "tcp://localhost:{port}".format(port=port)

def distributed_init(cfg: Config):
    if dist.is_available() and dist.is_initialized():
        warnings.warn(
            "Distributed is already initialized, cannot initialize twice!"
        )
    else:
        logger.info(
            "distributed init (rank {}): {}".format(
                cfg.distributed_training.distributed_rank,
                cfg.distributed_training.distributed_init_method
            )
        )
        dist.init_process_group(
            backend="nccl",
            init_method=cfg.distributed_training.distributed_init_method,
            world_size=cfg.distributed_training.distributed_world_size,
            rank=cfg.distributed_training.distributed_rank
        )
        logger.info(
            "initialized host {} as rank {}".format(
                socket.gethostname(),
                cfg.distributed_training.distributed_rank
            )
        )

        #perform a dummy all-reduce to initialize the NCCL communicator
        if torch.cuda.is_available():
            dist.all_reduce(torch.zeros(1).cuda())

    cfg.distributed_training.distributed_rank = dist.get_rank()

    if is_master(cfg.distributed_training):
        logging.getLogger().setLevel(logging.INFO)
    else:
        logging.getLogger().setLevel(logging.WARNING)

    return cfg.distributed_training.distributed_rank

def distributed_main(i, main, cfg: Config, kwargs):
    cfg.distributed_training.device_id = i
    if torch.cuda.is_available():
        torch.cuda.set_device(cfg.distributed_training.device_id)
    if cfg.distributed_training.distributed_rank is None:
        cfg.distributed_training.distributed_rank = kwargs.pop("start_rank", 0) + i
    
    cfg.distributed_training.distributed_rank = distributed_init(cfg)

    main(cfg, **kwargs)

    if dist.is_initialized():
        dist.barrier(get_global_group())

def call_main(cfg: Config, main, **kwargs):
    if (
        cfg.distributed_training.distributed_world_size > 1
        and cfg.distributed_training.distributed_init_method is None
    ):
        infer_init_method(cfg.distributed_training)
    
    if cfg.distributed_training.distributed_init_method is not None:
        start_rank = cfg.distributed_training.distributed_rank
        cfg.distributed_training.distributed_rank = None # assign automatically
        kwargs["start_rank"] = start_rank
        torch.multiprocessing.spawn(
            fn=distributed_main,
            args=(main, cfg, kwargs),
            nprocs=min(torch.cuda.device_count(), cfg.distributed_training.distributed_world_size),
            join=True
        )
    else:
        main(cfg, **kwargs)

def get_rank(group):
    return dist.get_rank(group=group)

def get_world_size(group):
    if dist.is_initialized():
        return dist.get_world_size(group=group)
    else:
        return 1

def get_global_group():
    if dist.is_initialized():
        if not hasattr(get_global_group, "_global_group"):
            # ideally we could use torch.distributed.group.WORLD, but it seems
            # to cause random NCCL hangs in some cases
            get_global_group._global_group = dist.new_group()
        return get_global_group._global_group
    else:
        return None

def get_global_rank():
    if dist.is_initialized():
        return dist.get_rank()
    else:
        return 0

def get_global_world_size():
    if dist.is_initialized():
        return dist.get_world_size()
    else:
        return 1

def get_data_parallel_group():
    """Get the data parallel group the caller rank belongs to."""
    return get_global_group()

def get_data_parallel_rank():
    """Return my rank for the data parallel group."""
    return get_rank(get_data_parallel_group())

def get_data_parallel_world_size():
    """Return world size for the data parallel group."""
    return get_world_size(get_data_parallel_group())

def all_reduce(tensor, group, op="sum"):
    if op == "sum":
        op = dist.ReduceOp.SUM
    elif op == "max":
        op = dist.ReduceOp.MAX
    else:
        raise NotImplementedError
    
    dist.all_reduce(tensor, op = op, group = group)
    
    return tensor

def all_gather_list(data, group=None, max_size=32768):
    """Gathers arbitrary data from all nodes into a list.

    Similar to :func:`~torch.distributed.all_gather` but for arbitrary Python
    data. Note that *data* must be picklable and any CUDA tensors will be moved
    to CPU and returned on CPU as well.

    Args:
        data (Any): data from the local worker to be gathered on other workers
        group: group of the collective
        max_size (int, optional): maximum size of the data to be gathered
            across workers
    """
    import utils.utils as utils

    if group is None:
        group = get_global_group()
    rank = get_rank(group = group)
    world_size = get_world_size(group = group)

    buffer_size = max_size * world_size
    if (
        not hasattr(all_gather_list, "_buffer")
        or all_gather_list._buffer.numel() < buffer_size
    ):
        all_gather_list._buffer = torch.cuda.ByteTensor(buffer_size)
        all_gather_list._cpu_buffer = torch.ByteTensor(max_size).pin_memory()
    buffer = all_gather_list._buffer
    buffer.zero_()
    cpu_buffer = all_gather_list._cpu_buffer

    data = utils.move_to_cpu(data)
    enc = pickle.dumps(data)
    enc_size = len(enc)
    header_size = 4 # size of header that contains the length of the encoded data
    size = header_size + enc_size
    if size > max_size:
        raise ValueError(
            "encoded data size ({}) exceeds max_size ({})".format(size, max_size)
        )
    
    header = struct.pack(">I", enc_size)
    cpu_buffer[:size] = torch.ByteTensor(list(header + enc))
    start = rank * max_size
    buffer[start : start + size].copy_(cpu_buffer[:size])
    
    all_reduce(buffer, group = group)

    buffer = buffer.cpu()
    try:
        result = []
        for i in range(world_size):
            out_buffer = buffer[i * max_size : (i + 1) * max_size]
            (enc_size,) = struct.unpack(">I", bytes(out_buffer[:header_size].tolist()))
            if enc_size > 0:
                result.append(
                    pickle.loads(
                        bytes(out_buffer[header_size : header_size + enc_size].tolist())
                    )
                )
        return result
    except pickle.UnpicklingError:
        raise Exception(
            "Unable to unpickle data from other workers. all_gather_list requires all "
            "workers to enter the function together, so this error usually indicates "
            "that the workers have fallen out of sync somehow. Workers can fall out of "
            "sync if one of them runs out of memory, or if there are other conditions "
            "in your training script that can cause one worker to finish an epoch "
            "while other workers are still iterating over their portions of the data. "
            # "Try rerunning with --ddp-backend=legacy_ddp and see if that helps."
        )

def batch_all_gather(tensor, group, return_tensor=False):
    """Perform an all-gather operation considering tensors with different batch size"""
    world_size = get_world_size(group=group)
    rank = get_rank(group=group)

    size_list = [
        tensor.new_zeros(tensor.dim(), dtype=torch.int64) for _ in range(world_size)
    ]
    local_size = tensor.new_tensor(tensor.shape, dtype=torch.int64)
    dist.all_gather(size_list, local_size, group=group)

    max_size = torch.stack(size_list).max(dim=0)[0][0]
    size_offsets = [max_size - size[0] for size in size_list]

    if local_size[0] != max_size:
        offset = torch.cat(
            (
                tensor.new_tensor([max_size - local_size[0]]),
                local_size[1:]
            )
        )
        padding = tensor.new_zeros(tuple(int(dim) for dim in offset), dtype=torch.uint8)
        tensor = torch.cat((tensor, padding), dim=0)

    tensor_list = [
        tensor if i == rank else torch.empty_like(tensor) for i in range(world_size)
    ]
    dist.all_gather(tensor_list, tensor, group=group)
    tensor_list = [
        tensor[:max_size-size_offsets[i]] for i, tensor in enumerate(tensor_list)
    ]
    if return_tensor:
        return torch.stack(tensor_list, dim=0)
    else:
        return tensor_list

def all_gather(tensor, group, return_tensor=False):
    """Perform an all-gather operation."""
    world_size = get_world_size(group = group)
    rank = get_rank(group = group)
    tensor_list = [
        tensor if i == rank else torch.empty_like(tensor) for i in range(world_size)
    ]
    dist.all_gather(tensor_list, tensor, group = group)
    if return_tensor:
        return torch.stack(tensor_list, dim = 0)
    else:
        return tensor_list

def broadcast(tensor, src, group):
    dist.broadcast(tensor, src=src, group=group)

def broadcast_tensors(
    tensors: Optional[List[torch.Tensor]],
    src_rank: int,
    group: object,
    dist_device: Optional[torch.device] = None
) -> List[torch.Tensor]:
    """
    Broadcast a list of tensors without other (non-src) ranks needing to know
    the dtypes/shapes of the tensors.
    """
    if dist_device is None:
        dist_device = torch.device("cuda")
    
    #share metadata first to simplify transfer
    is_src_rank = (get_rank(group) == src_rank)
    if is_src_rank:
        metadata = [
            {"size": t.size(), "dtype": t.dtype, "device": t.device} for t in tensors
        ]
        metadata = _broadcast_object_slow(metadata, src_rank, group, dist_device)
    else:
        metadata = _broadcast_object_slow(None, src_rank, group, dist_device)
    
    out_tensors = []
    for i, meta in enumerate(metadata):
        if is_src_rank:
            tensor = tensors[i]
            broadcast(tensors[i].to(dist_device), src=src_rank, group=group)
        else:
            tensor = torch.zeros(
                [meta["size"].numel()], dtype=meta["dtype"], device=dist_device
            )
            broadcast(tensor, src=src_rank, group=group)
        tensor = tensor.view(meta["size"]).to(meta["device"])
        out_tensors.append(tensor)    
    return out_tensors

def broadcast_object(
    obj : Any,
    src_rank : int,
    group : object,
    dist_device: Optional[torch.device] = None
) -> Any:
    """Broadcast an arbitrary Python object to other workers."""
    if dist_device is None:
        dist_device = torch.device("cuda")
    
    if get_rank(group) == src_rank:
        # split the tensors from the non-tensors so we can broadcast them
        # directly, avoiding unnecessary serialization/deserialization
        tensors = []
        obj = _split_tensors_from_obj(obj, tensors)
        obj = _broadcast_object_slow(obj, src_rank, group, dist_device)
        tensors = broadcast_tensors(tensors, src_rank, group, dist_device)
    else:
        obj = _broadcast_object_slow(None, src_rank, group, dist_device)
        tensors = broadcast_tensors(None, src_rank, group, dist_device)
    return _put_tensors_in_obj(obj, tensors)

def _broadcast_object_slow(
    obj: Any, src_rank: int, group: object, dist_device: torch.device,
) -> Any:
    rank = get_rank(group)
    if get_rank(group) == src_rank:
        # Emit data
        buffer = io.BytesIO()
        torch.save(obj, buffer)
        buffer = torch.ByteTensor(buffer.getbuffer()).to(dist_device)
        length = torch.LongTensor([len(buffer)]).to(dist_device)
        broadcast(length, src=src_rank, group=group)
        broadcast(buffer, src=src_rank, group=group)
    else:
        # Fetch from the source
        length = torch.LongTensor([0]).to(dist_device)
        broadcast(length, src=src_rank, group=group)
        buffer = torch.ByteTensor(int(length.item())).to(dist_device)
        broadcast(buffer, src=src_rank, group=group)
        buffer = io.BytesIO(buffer.cpu().numpy())
        obj = torch.load(buffer, map_location="cpu")
    return obj

@dataclass(frozen=True)
class _TensorPlaceholder:
    index: int

def _split_tensors_from_obj(obj: Any, tensors: List[torch.Tensor]) -> Any:
    if torch.is_tensor(obj):
        placeholder = _TensorPlaceholder(index = len(tensors))
        tensors.append(obj)
        return placeholder
    elif isinstance(obj, dict):
        return {k: _split_tensors_from_obj(v, tensors) for k,v in obj.items()}
    elif isinstance(obj, list):
        return [_split_tensors_from_obj(v, tensors) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(_split_tensors_from_obj(v, tensors) for v in obj)
    elif isinstance(obj, set):
        return {_split_tensors_from_obj(v, tensors) for v in obj}
    else:
        return obj

def _put_tensors_in_obj(obj: Any, tensors: List[torch.Tensor]) -> Any:
    if isinstance(obj, _TensorPlaceholder):
        return tensors[obj.index]
    elif isinstance(obj, dict):
        return {k: _put_tensors_in_obj(v, tensors) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_put_tensors_in_obj(v, tensors) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(_put_tensors_in_obj(v, tensors) for v in obj)
    elif isinstance(obj, set):
        return {_put_tensors_in_obj(v, tensors) for v in obj}
    else:
        return obj


import logging
from dataclasses import dataclass, field
from typing import Tuple

import torch
import torch.nn as nn

from genhpf.models import register_model
from genhpf.models.genhpf import GenHPF, GenHPFConfig
from genhpf.modules import GradMultiply, GumbelVectorQuantizer, LayerNorm
from genhpf.utils import utils
from genhpf.utils.data_utils import compute_mask_indices

logger = logging.getLogger(__name__)


@dataclass
class GenHPFWav2Vec2Config(GenHPFConfig):
    logit_temp: float = field(default=0.1, metadata={"help": "temperature to divide logits by"})
    latent_vars: int = field(
        default=320, metadata={"help": "number of latent variables V in each group of the codebook"}
    )
    latent_groups: int = field(
        default=2, metadata={"help": "number of groups G of latent variables in the codebook"}
    )
    latent_temp: Tuple[float, float, float] = field(
        default=(2, 0.5, 0.999995),
        metadata={
            "help": "temperature for latent variable sampling. "
            "can be tuple of 3 values (start, end, decay)"
        },
    )
    final_dim: int = field(
        default=128, metadata={"help": "project final representations and targets to this many dimensions."}
    )

    num_negatives: int = field(
        default=25, metadata={"help": "number of negative examples from the same sample"}
    )
    codebook_negatives: int = field(default=0, metadata={"help": "number of negative examples codebook"})

    # mask
    mask_prob: float = field(default=0.65, metadata={"help": "probability of replacing a token with mask"})
    mask_length: int = field(default=1, metadata={"help": "mask length"})
    no_mask_overlap: bool = field(default=False, metadata={"help": "whether to allow masks to overlap"})
    mask_min_space: int = field(
        default=0, metadata={"help": "min space between spans (if no overlap is enabled)"}
    )

    feature_grad_mult: float = field(
        default=0.1, metadata={"help": "multiply event encoder gradients by this"}
    )
    dropout_input: float = field(default=0.1, metadata={"help": "dropout to apply to the input"})
    dropout_features: float = field(default=0.1, metadata={"help": "dropout to apply to the features"})


@register_model("genhpf_wav2vec2", dataclass=GenHPFWav2Vec2Config)
class GenHPFWav2Vec2(GenHPF):
    def __init__(self, cfg: GenHPFWav2Vec2Config):
        super().__init__(cfg)

        self.logit_temp = cfg.logit_temp

        self.mask_prob = cfg.mask_prob
        self.mask_length = cfg.mask_length
        self.no_mask_overlap = cfg.no_mask_overlap
        self.mask_min_space = cfg.mask_min_space

        self.feature_grad_mult = cfg.feature_grad_mult

        self.codebook_negatives = cfg.codebook_negatives

        self.quantizer = GumbelVectorQuantizer(
            dim=cfg.agg_embed_dim,
            num_vars=cfg.latent_vars,
            temp=cfg.latent_temp,
            groups=cfg.latent_groups,
            combine_groups=False,
            vq_dim=cfg.agg_embed_dim,
            time_first=True,
        )

        self.layer_norm = LayerNorm(cfg.agg_embed_dim)
        self.dropout_input = nn.Dropout(cfg.dropout_input)
        self.dropout_features = nn.Dropout(cfg.dropout_features)
        self.project_q = nn.Linear(cfg.agg_embed_dim, cfg.final_dim)
        self.final_proj = nn.Linear(cfg.agg_embed_dim, cfg.final_dim)
        self.mask_emb = nn.Parameter(torch.FloatTensor(cfg.agg_embed_dim).uniform_())
        self.cross_sample_negatives = 0
        self.n_negatives = cfg.num_negatives

    @classmethod
    def build_model(cls, cfg):
        """Build a new model instance."""
        return cls(cfg)

    def compute_preds(self, x, y, negatives):
        neg_is_pos = (y == negatives).all(-1)
        y = y.unsqueeze(0)
        targets = torch.cat([y, negatives], dim=0)

        logits = torch.cosine_similarity(x.float(), targets.float(), dim=-1)

        logits = logits / self.logit_temp
        logits = logits.type_as(x)

        if neg_is_pos.any():
            logits[1:][neg_is_pos] = float("-inf")

        return logits

    def apply_mask(
        self,
        x: torch.Tensor,
        padding_mask: torch.Tensor,
        mask_indices=None,
    ):
        bsz, tsz, csz = x.shape

        if self.mask_prob > 0:
            if mask_indices is None:
                mask_indices = compute_mask_indices(
                    shape=(bsz, tsz),
                    padding_mask=padding_mask,
                    mask_prob=self.mask_prob,
                    mask_length=self.mask_length,
                    mask_type="static",
                    min_masks=2,
                    no_overlap=self.no_mask_overlap,
                    min_space=self.mask_min_space,
                )
                mask_indices = torch.from_numpy(mask_indices).to(x.device)
            x[mask_indices] = self.mask_emb
        else:
            mask_indices = None

        return x, mask_indices

    def sample_negatives(self, y, num):
        if self.n_negatives == 0 and self.cross_sample_negatives == 0:
            return y.new(0)

        batch_size, time_size, feature_size = y.shape
        y = y.view(-1, feature_size)  # B x T x C -> (B x T) x C

        cross_high = time_size * batch_size
        high = time_size
        with torch.no_grad():
            assert high > 1, f"{batch_size, time_size, feature_size}"

            if self.n_negatives > 0:
                time_sizes = utils.buffered_arange(num).unsqueeze(-1).expand(-1, self.n_negatives).flatten()
                neg_idxs = torch.randint(low=0, high=high - 1, size=(batch_size, self.n_negatives * num))
                neg_idxs[neg_idxs >= time_sizes] += 1

            if self.cross_sample_negatives > 0:
                time_sizes = (
                    utils.buffered_arange(num).unsqueeze(-1).expand(-1, self.cross_sample_negatives).flatten()
                )

                cross_neg_idxs = torch.randint(
                    low=0, high=cross_high - 1, size=(batch_size, self.cross_sample_negatives * num)
                )
                cross_neg_idxs[cross_neg_idxs >= time_sizes] += 1

        if self.n_negatives > 0:
            for i in range(1, batch_size):
                neg_idxs[i] += i * high
        else:
            neg_idxs = cross_neg_idxs

        if self.cross_sample_negatives > 0 and self.n_negatives > 0:
            neg_idxs = torch.cat([neg_idxs, cross_neg_idxs], dim=1)

        negs = y[neg_idxs.view(-1)]
        negs = negs.view(
            batch_size, num, self.n_negatives + self.cross_sample_negatives, feature_size
        ).permute(
            2, 0, 1, 3
        )  # to N x B x T x C

        return negs, neg_idxs

    def get_logits(self, sample, net_output):
        logits = net_output["x"]
        logits = logits.transpose(0, 2)
        logits = logits.reshape(-1, logits.size(-1))

        return logits

    def get_targets(self, sample, net_output):
        x = net_output["x"]
        return x.new_zeros(x.size(1) * x.size(2), dtype=torch.long)

    def get_extra_losses(self, net_output):
        pen = []

        if "prob_perplexity" in net_output:
            pen.append((net_output["num_vars"] - net_output["prob_perplexity"]) / net_output["num_vars"])

        if "features_pen" in net_output:
            pen.append(net_output["features_pen"])

        return pen

    def forward(
        self,
        input_ids: torch.Tensor,
        type_ids: torch.Tensor = None,
        dpe_ids: torch.Tensor = None,
        padding_mask: torch.Tensor = None,
        **kwargs,
    ):
        if self.feature_grad_mult > 0:
            features_ret = super().forward(
                input_ids=input_ids,
                type_ids=type_ids,
                dpe_ids=dpe_ids,
                padding_mask=padding_mask,
                encoder_only=True,
                **kwargs,
            )
            features = features_ret["x"]
            padding_mask = features_ret["padding_mask"]
            if self.feature_grad_mult != 1.0:
                features = GradMultiply.apply(features, self.feature_grad_mult)
        else:
            with torch.no_grad():
                features = super().forward(
                    input_ids=input_ids,
                    type_ids=type_ids,
                    dpe_ids=dpe_ids,
                    padding_mask=padding_mask,
                    encoder_only=True,
                    **kwargs,
                )
                features = features_ret["x"]
                padding_mask = features_ret["padding_mask"]

        features_pen = features.float().pow(2).mean()

        features = self.layer_norm(features)
        unmasked_features = features.clone()

        features = self.dropout_input(features)
        unmasked_features = self.dropout_features(unmasked_features)

        num_vars = None
        code_ppl = None
        prob_ppl = None
        curr_temp = None

        if padding_mask is None:
            padding_mask = input_ids[:, :, 1].eq(0).to(features.device)

        x, mask_indices = self.apply_mask(features, padding_mask, mask_indices=None)
        if mask_indices is not None:
            y = unmasked_features[mask_indices].view(
                unmasked_features.size(0), -1, unmasked_features.size(-1)
            )
        else:
            y = unmasked_features

        x = self.event_aggregator(x, src_key_padding_mask=padding_mask)

        q = self.quantizer(y, produce_targets=False)
        y = q["x"]
        num_vars = q["num_vars"]
        code_ppl = q["code_perplexity"]
        prob_ppl = q["prob_perplexity"]
        curr_temp = q["temp"]

        y = self.project_q(y)

        negs, _ = self.sample_negatives(y, y.size(1))

        if self.codebook_negatives > 0:
            cb_negs = self.quantizer.sample_from_codebook(y.size(0) * y.size(1), self.codebook_negatives)
            cb_negs = cb_negs.view(self.codebook_negatives, y.size(0), y.size(1), -1)
            cb_negs = self.project_q(cb_negs)
            negs = torch.cat([negs, cb_negs], dim=0)

        x = x[mask_indices].view(x.size(0), -1, x.size(-1))

        x = self.final_proj(x)
        x = self.compute_preds(x, y, negs)

        output = {"x": x, "features_pen": features_pen, "mask_indices": mask_indices}
        if prob_ppl is not None:
            output["prob_perplexity"] = prob_ppl
            output["code_perplexity"] = code_ppl
            output["num_vars"] = num_vars
            output["temp"] = curr_temp

        return output

    def get_pretraining_parameter_names(self):
        ret = []
        ret.append("mask_emb")
        ret.extend(["quantizer" + "." + x[0] for x in self.quantizer.named_parameters()])
        ret.extend(["layer_norm" + "." + x[0] for x in self.layer_norm.named_parameters()])
        ret.extend(["project_q" + "." + x[0] for x in self.project_q.named_parameters()])
        ret.extend(["final_proj" + "." + x[0] for x in self.final_proj.named_parameters()])
        return ret



import argparse
import os
import random

import h5pickle
from tqdm import tqdm


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "root", metavar="DIR", type=str, help="root directory containing the data.h5 and label.csv"
    )
    parser.add_argument("--dest", default=".", type=str, metavar="DIR", help="output directory")
    parser.add_argument("--prefix", default="", type=str, help="prefix for the output files")
    parser.add_argument(
        "--valid-percent",
        default=0.1,
        type=float,
        metavar="D",
        help="percentage of data to use as validation and test set (between 0 and 0.5)",
    )
    parser.add_argument("--seed", default=42, type=int, metavar="N", help="random seed")
    return parser


def main(args):
    assert 0 <= args.valid_percent <= 0.5, (
        f"Invalid valid-percent: {args.valid_percent}. " "Must be between 0 and 0.5."
    )

    if len(args.prefix) > 0 and not args.prefix.endswith("_"):
        args.prefix += "_"

    root_path = os.path.realpath(args.root)
    data_path = os.path.join(root_path, "data.h5")
    label_path = os.path.join(root_path, "label.csv")
    rand = random.Random(args.seed)

    if not os.path.exists(args.dest):
        os.makedirs(args.dest)

    with (
        open(os.path.join(args.dest, f"{args.prefix}train.tsv"), "w") as train_f,
        open(os.path.join(args.dest, f"{args.prefix}valid.tsv"), "w") as valid_f,
        open(os.path.join(args.dest, f"{args.prefix}test.tsv"), "w") as test_f,
    ):
        print(data_path, file=train_f)
        print(label_path, file=train_f)
        print(data_path, file=valid_f)
        print(label_path, file=valid_f)
        print(data_path, file=test_f)
        print(label_path, file=test_f)

        def write(subjects, dest, split=None):
            for subject in tqdm(subjects, total=len(subjects), desc=split):
                print(subject, file=dest)

        data = h5pickle.File(data_path, "r")["ehr"]
        subjects = list(data.keys())
        rand.shuffle(subjects)

        valid_len = int(len(subjects) * args.valid_percent)
        test_len = int(len(subjects) * args.valid_percent)
        train_len = len(subjects) - valid_len - test_len

        train = subjects[:train_len]
        valid = subjects[train_len : train_len + valid_len]
        test = subjects[train_len + valid_len :]

        write(train, train_f, split=f"{args.prefix}train")
        write(valid, valid_f, split=f"{args.prefix}valid")
        write(test, test_f, split=f"{args.prefix}test")


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    main(args)
